import argparse
from datetime import datetime, timedelta
import logging
import os
import shutil

from django.conf import settings
from django.core.management.base import BaseCommand
from django.core.exceptions import ImproperlyConfigured

from django_b2.storage import B2Storage, LOGFILE_PREFIX, LOG_LOCATION


HISTORY_LOCATION = 'history/'


log = logging.getLogger('avis')


class Command(BaseCommand):
    help = ("Delete local media (copy of backblaze upload) and copy log to <MEDIA_ROOT>/%s/%s. "
           "This can be used together with the B2_LOCAL_MEDIA='ML' setting." % (LOG_LOCATION, HISTORY_LOCATION))

    def add_arguments(self, parser):
        def positive_type(x):
            x = int(x)
            if x < 1:
                raise argparse.ArgumentTypeError("Minimum for --days or --hours is 1.")
            return x

        parser.add_argument('--days', type=positive_type,
                            help="Delete local media older than N days. """
                                 "Default is 1 day used if both --days and --hours are missing")
        parser.add_argument('--hours', type=positive_type)
        parser.add_argument('--no-history', action='store_true',
                            help="Do not backup solved log files to <MEDIA_ROOT>/%s/%s"
                            % (LOG_LOCATION, HISTORY_LOCATION))

    def handle(self, *args, **options):
        if not settings.MEDIA_ROOT:
            raise ImproperlyConfigured("Cannot seek for local media files, MEDIA_ROOT isn't set.")
        log_dir = os.path.join(settings.MEDIA_ROOT, LOG_LOCATION)
        if not os.path.isdir(log_dir):
            raise ImproperlyConfigured("Cannot seek for local media files, MEDIA_ROOT isn't set.")

        hours = options.get('hours', 0)
        if options['days']:
            hours += 24 * options['days']
        elif not hours:
            hours = 24  # default is 1 day
        self.handle_clean(hours, log_dir, no_history=options['no_history'])

    def handle_clean(self, hours, log_dir, no_history=False):
        bk_dir = None
        if not no_history:
            bk_dir = os.path.join(log_dir, HISTORY_LOCATION)
            os.makedirs(bk_dir, exist_ok=True)
        max_filename = self.get_max_filename(hours)
        for log_file in os.listdir(log_dir):
            if log_file.startswith(LOGFILE_PREFIX) and log_file <= max_filename:
                log_full = os.path.join(log_dir, log_file)
                if os.path.isfile(log_full):  # directories, like history/ shouldn't start with LOGFILE_PREFIX, but..
                    log_ok = self.handle_clean_logfile(log_full)
                    if log_ok:
                        if bk_dir is None:
                            os.remove(log_full)
                        else:
                            shutil.move(log_full, os.path.join(bk_dir, log_file))  # move log file to history/

    def handle_clean_logfile(self, log_full):
        log_ok = True
        with open(log_full, 'r') as f:
            media_files = f.readlines()
            for media_file in media_files:
                if media_file:
                    media_file = os.path.join(settings.MEDIA_ROOT, media_file)
                    if os.path.isfile(media_file):
                        try:
                            os.remove(media_file)
                        except Exception:
                            log_ok = False
        return log_ok

    def get_max_filename(self, hours):
        assert hours >= 1, "Expected to delete 1+ hours old media files, but hours < 1; stopping, to be safe."
        return B2Storage.get_logfile_name(datetime.now() - timedelta(hours=hours + 1))  # N hours -> interval(N, N+1)



"""
    B2Storage
        uses b2sdk

    add this to your Django settings:
        DEFAULT_FILE_STORAGE = 'django_b2.storage.B2Storage'
        B2_APP_KEY_ID = '000xxxxxxxxxxxx000000000n'
        B2_APP_KEY = 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'
        B2_BUCKET_NAME = '<bucket-name>'

    this storage makes names unique by change path/name into path/<uuid>/name
        so it is not compatible with storage uploaded using django-backblazeb2-storage
            where files with same names are uploaded without name change and as versions of same file

    behaviour of exceptions (FileNotFoundError, NotADirectoryError, ..) is not fully compatible

    methods _open, _save works but need experienced revision and rewrite (please help)
"""


import importlib
from io import BytesIO
import os
from datetime import datetime

from django.conf import settings
from django.core.files.storage import FileSystemStorage, Storage
from django.core.files.base import ContentFile, File
from django.utils.deconstruct import deconstructible

from django_b2.backblaze_b2 import BackBlazeB2


# B2File related imports, TODO: mix with non-experimental imports if B2File will be used
from shutil import copyfileobj
from tempfile import SpooledTemporaryFile


# if 'L' in settings.B2_LOCAL_MEDIA, list (log) of uploaded files will be created in MEDIA_ROOT/<LOG_LOCATION>
LOG_LOCATION = '_log'
LOGFILE_PREFIX = 'upload_'


# experimental (B2File class, B2Storage._open), inspired from django-storages DropBoxFile
# TODO: this works, but need revision and rewrite, please help!
@deconstructible
class B2File(File):
    def __init__(self, name, download_dest):
        self.name = name
        self._download_dest = download_dest
        self._file = None

    def _get_file(self):
        if self._file is None:
            self._file = SpooledTemporaryFile()
            with BytesIO(self._download_dest.get_bytes_written()) as file_content:
                copyfileobj(file_content, self._file)
            self._file.seek(0)
            # self._download_dest = None
        return self._file

    def _set_file(self, value):
        self._file = value

    file = property(_get_file, _set_file)


standard_mode = True
if 'B2Storage' not in settings.DEFAULT_FILE_STORAGE:
    """
        not sure if this could work properly but the goal is to make following in tests possible:
            from django.test import override_settings
            @override_settings(DEFAULT_FILE_STORAGE='django.core.files.storage.FileSystemStorage')
        see https://github.com/pyutil/django-b2/issues/4
    """
    dfs = settings.DEFAULT_FILE_STORAGE.rsplit('.', 1)  # ['django.core.files.storage', 'FileSystemStorage']
    if len(dfs) == 2:
        standard_mode = False
        module_name, class_name = dfs
        module = importlib.import_module(module_name)
        B2Storage = getattr(module, class_name)


if standard_mode:
    """
        standard working mode, with B2Storage
    """
    @deconstructible
    class B2Storage(Storage):
        location = ''   # required at least for django-tenant-schemas
        _b2 = None
        _fs = None
        _log = None

        @property
        def b2(self):
            self.check_init()
            return self._b2

        @property
        def fs(self):
            self.check_init()
            return self._fs

        @property
        def log(self):
            self.check_init()
            return self._log

        def check_init(self):
            # we always check for _b2, because _fs & _log can stay None after lazy_init
            if self._b2 is None:
                self.lazy_init()

        def lazy_init(self):
            application_key_id = settings.B2_APP_KEY_ID
            application_key = settings.B2_APP_KEY
            bucket_name = settings.B2_BUCKET_NAME

            self._b2 = BackBlazeB2()
            self.authorize(application_key_id, application_key)
            self.set_bucket(bucket_name)

            if hasattr(settings, 'B2_LOCAL_MEDIA'):
                assert settings.MEDIA_ROOT, 'B2_LOCAL_MEDIA used. Please set MEDIA_ROOT in your settings.'
                if 'M' in settings.B2_LOCAL_MEDIA:
                    self._fs = FileSystemStorage()
                if 'L' in settings.B2_LOCAL_MEDIA:
                    self._log = os.path.join(settings.MEDIA_ROOT, LOG_LOCATION)
                    os.makedirs(self._log, exist_ok=True)

        # you can re-authorize later
        def authorize(self, application_key_id, application_key):
            return self.b2.authorize("production", application_key_id, application_key)

        # you can change bucket later
        def set_bucket(self, bucket_name):
            return self.b2.set_bucket(bucket_name)

        # ------------- django Storage extension --------------

        # method added with regard to Django 3.0 compatibility, both (available/alternative) do the same: add <uuid>/
        def get_alternative_name(self, file_root, file_ext=''):
            name = '%s%s' % (file_root, file_ext)
            return self.b2.get_alternative_name(name)

        # both (available/alternative) do the same: add <uuid>/
        # important to have this: to prevent inefficient .exists() call
        def get_available_name(self, name, max_length=None):
            return self.b2.get_alternative_name(name)

        def _open(self, name, mode='rb'):
            '''
            # this is based on django-backblazeb2-storage
            # TODO: this works, but need revision and rewrite, please help!
            resp = self.b2.download_file(name)
            output = BytesIO()
            output.write(resp)
            output.seek(0)
            return File(output, name)
            '''
            if self.fs is not None and os.path.isfile(os.path.join(settings.MEDIA_ROOT, name)):
                try:
                    return self.fs._open(name)
                except Exception:
                    pass
            # this is inspired by django-storage:DropBox
            # TODO: this works, but need revision and rewrite, please help!
            download_dest = self.b2.download_file_download_dest(name)
            return B2File(name, download_dest)

        def _save(self, name, f, max_length=None):
            if self.fs is not None:
                self.fs.save(name, ContentFile(f.read()))
                f.seek(0)
            if self.log is not None:
                log_file = self.get_logfile_name(datetime.now())
                log_file = os.path.join(self.log, log_file)
                with open(log_file, 'a') as lf:
                    lf.writelines([name + '\n'])
            response = self.b2.upload_file(name, f)
            return response.file_name

        @staticmethod
        def get_logfile_name(dt):
            return LOGFILE_PREFIX + dt.isoformat()[:13]

        def delete(self, name):
            return self.b2.delete_by_name(name)

        # maybe expensive
        def listdir(self, path):
            return self.b2.listdir(path)

        def exists(self, name):
            return self.b2.file_id_by_name(name) is not None

        def size(self, name):
            return self.b2.get_file_info_by_name(name)['contentLength']

        def url(self, name):
            return self.b2.get_download_url(name)

        def xxxpath(self, name):
            # this was here but seems this is problem because at least Wagtail determinates by retval if the file is stored
            #  locally, so it should (seems) return a full path not the relative one
            #    # This is needed because Django will throw an exception if it's not
            #    # overridden by Storage subclasses. We don't need it.
            #    return name
            '''
            """
            Return a local filesystem path where the file can be retrieved using
            Python's built-in open() function. Storage systems that can't be
            accessed using open() should *not* implement this method.
            """
            raise NotImplementedError("This backend doesn't support absolute paths.")
            '''

            full = os.path.join(settings.MEDIA_ROOT, name)
            if os.path.isfile(full):
                return full
            return super().path(name)

        def get_accessed_time(self, name):
            return self.b2.get_accessed_time(name, use_tz=settings.USE_TZ)

        def get_created_time(self, name):
            return self.b2.get_created_time(name, use_tz=settings.USE_TZ)

        def get_modified_time(self, name):
            return self.b2.get_modified_time(name, use_tz=settings.USE_TZ)



# Tenant aware storages
# Can be used together with postgres and django-tenant-schemas..
# ..and probably (need testing) with django-tenants if you add django-tenant-schemas into requirements (but not to INSTALLED_APPS)

# https://github.com/bernardopires/django-tenant-schemas/issues/565

from tenant_schemas.storage import TenantStorageMixin

# original: DEFAULT_FILE_STORAGE = 'django_b2.storage.B2Storage'
from django_b2.storage import B2Storage


# DEFAULT_FILE_STORAGE = 'django_b2.tenant_storage.TenantB2Storage'
class TenantB2Storage(TenantStorageMixin, B2Storage):
    pass



# except of its standard usage this can be used as the copy/upload command to backblaze
#   for help write: python backblaze_b2 --help
#   warning: additional imports for this usage are in 'command' section bellow

from datetime import datetime
import uuid

from b2sdk.v1 import B2Api, DownloadDestBytes, InMemoryAccountInfo, UploadSourceBytes

try:
    from django.utils import timezone
    USE_TZ = True
except (ModuleNotFoundError, ImportError) as exc:
    USE_TZ = False


LS_FETCH_COUNT = 1000    # probably 100 is default, 1000-10000 are paid based on 1000-chunks
DOWNLOAD_TIMEOUT = 1800  # validity of download links [s]


class BackBlazeB2(object):
    _b2_api = None
    bucket = None
    authorized = False

    @property
    def b2_api(self):
        if self._b2_api is None:
            info = InMemoryAccountInfo()
            self._b2_api = B2Api(info)
        return self._b2_api

    # need to call this after instantiate; you could change this later
    def authorize(self, name, application_key_id, application_key):
        # name: usually can be: "production"
        self.b2_api.authorize_account(name, application_key_id, application_key)
        self.authorized = True

    # need to call this after instantiate; you could change this later
    def set_bucket(self, bucket_name):
        self.bucket = self.b2_api.get_bucket_by_name(bucket_name)

    def is_prepared(self):   # was .authorize and .set_bucket called earlier?
        return self.authorized and self.bucket is not None

    # ------------------------------------------------------------------

    def download_file(self, name):                 # return DownloadDestBytes() str content
        return self.download_file_download_dest(name).get_bytes_written()

    def download_file_download_dest(self, name):   # return DownloadDestBytes() object
        assert self.is_prepared()
        download_dest = DownloadDestBytes()
        self.bucket.download_file_by_name(self.fix_b2name(name), download_dest)
        return download_dest

    def upload_file(self, name, f):
        assert self.is_prepared()
        content = f.read()
        # warning: 'InMemoryUploadedFile' object has no attribute 'mode' (and content read is b'..')
        if hasattr(f, 'mode') and 'b' not in f.mode:
            content = content.encode()
        uploadsource = UploadSourceBytes(content)
        return self.bucket.upload(uploadsource, self.fix_b2name(name))

    def upload_file_unique_name_outside_django(self, name, f):
        """
            django storage auto-calls get_alternative_name() so from django call upload_file() directly
        """
        return self.upload_file(self.get_alternative_name(name), f)

    def delete_by_name(self, name):
        for version in self.versions_by_name(name):
            self.delete_file_version(name, version['fileId'])

    def delete_file_version(self, name, file_id):
        assert self.is_prepared()
        return self.bucket.delete_file_version(file_id, self.fix_b2name(name))

    # maybe expensive
    def purge_bucket(self):
        # purge large uncommitted uploads is not implemented here
        while True:
            versions = self._list_versions()
            for version in versions['files']:
                self.delete_file_version(version['fileName'], version['fileId'])
            if versions['nextFileId'] is None:
                break

    def get_file_info(self, file_id):
        assert self.is_prepared()
        return self.b2_api.get_file_info(file_id)

    def get_file_info_by_name(self, name):
        assert self.is_prepared()
        file_id = self.file_id_by_name(name)
        return self.get_file_info(file_id)

    def get_download_url(self, name):
        assert self.is_prepared()
        name = self.fix_b2name(name)
        url = self.bucket.get_download_url(name)   # == self.b2.get_file_url(name)
        if 'private' in self.bucket.bucket_dict['bucketType'].lower():
            url += '?Authorization=' + self.bucket.get_download_authorization(name,
                                                                            valid_duration_in_seconds=DOWNLOAD_TIMEOUT)
        return url

    # maybe expensive
    def ls(self, path, fetch_count=LS_FETCH_COUNT):
        assert self.is_prepared()
        path = self.fix_b2name(path)
        # this is generator (which handles chunks, so it returns all entries & fetch_count has an internal effect only)
        return self.bucket.ls(folder_to_list=path, fetch_count=fetch_count)

    # maybe expensive
    def listdir(self, path):
        """
            :return: compatible with Django storage : listdir
        """
        path = self.fix_b2name(path)
        if len(path) and path[-1] != '/':
            path += '/'
        dirs = set()
        files = []
        for file_info, file_path in self.ls(path):
            if file_path is None:
                files.append(file_info.file_name.rsplit('/', 1)[-1])
            else:
                file_path = file_path[len(path):]
                dirs.add(file_path.split('/', 1)[0])
        return list(dirs), files

    def file_by_name(self, name):
        # is there a less stupid way to do it?
        assert self.is_prepared()
        name = self.fix_b2name(name)
        fn = self.b2_api.raw_api.list_file_names(self.b2_api.account_info.get_api_url(),
                                                 self.b2_api.account_info.get_account_auth_token(),
                                                 self.bucket.get_id(),
                                                 start_file_name=name, max_file_count=1, prefix=name)
        return self._x_by_name_result(fn, name)

    def versions_by_name(self, name):
        # is there a less stupid way to do it?
        name = self.fix_b2name(name)
        fn = self._list_versions(name)
        return self._x_by_name_result(fn, name)

    def _list_versions(self, name=None, start_file_name=None, start_file_id=None):
        # returns at most 1000 file names per transaction
        assert self.is_prepared()
        if name:
            start_file_name = name
            kwargs = {'prefix': name}
        else:
            kwargs = {}
        retval = self.b2_api.raw_api.list_file_versions(self.b2_api.account_info.get_api_url(),
                                                        self.b2_api.account_info.get_account_auth_token(),
                                                        self.bucket.get_id(),
                                                        start_file_name=start_file_name,
                                                        start_file_id=start_file_id,
                                                        max_file_count=LS_FETCH_COUNT,
                                                        **kwargs)
        if type(retval) == dict:
            return retval
        else:
            return {'files': []}  # seems at 2020-01-06 b2_api returns str:'files', which is probably a minor bug

    def _x_by_name_result(self, fn, name):
        res = []
        for f in fn['files']:
            if f['fileName'] == name:
                res.append(f)
        return res

    def file_id_by_name(self, name):
        # is there a less stupid way to do it?
        f = self.file_by_name(name)
        if len(f) == 0:
            return None
        else:
            return f[0]['fileId']

    def get_accessed_time(self, name, use_tz=USE_TZ):  # we haven't accessed time, so..
        return self.get_modified_time(name, use_tz)        # TODO: isn't NotImplemented better?

    def get_created_time(self, name, use_tz=USE_TZ):
        assert self.is_prepared()
        versions = self.versions_by_name(name)
        return self._get_time_from_fileinfo(versions, use_tz)

    def get_modified_time(self, name, use_tz=USE_TZ):
        assert self.is_prepared()
        f = self.file_by_name(name)
        return self._get_time_from_fileinfo(f, use_tz)

    @staticmethod
    def _get_time_from_fileinfo(fileinfo, use_tz):
        if not fileinfo:
            return None
        ts = int(fileinfo[-1]['uploadTimestamp'] / 1000)
        if use_tz:
            return datetime.utcfromtimestamp(ts).replace(tzinfo=timezone.utc)
        else:
            return datetime.fromtimestamp(ts)

    @staticmethod
    def fix_b2name(name):
        # https://github.com/pyutil/django-b2/issues/2
        # behaviour on Windows, thx to Sam Weaver, 24.5.2020, v0.5.5
        if "\\" in name:
            name = name.replace("\\", "/")

        # file name should be always relative to bucketl; with following we can save abs paths 1:1 to the bucket
        if name[:1] == '/':
            name = name[1:]

        return name

    # named compatible with Django 3.0
    def get_alternative_name(self, name):
        name = name.split('/')
        name.insert(len(name) - 1, str(uuid.uuid4()))
        return '/'.join(name)

    def get_original_name(self, name):
        return get_original_name(name)


# ------- utils, can be imported separately -----------------------


# revert get_alternative_name()
def get_original_name(name):
    name = name.split('/')
    if len(name) > 1:
        del name[-2]
    return '/'.join(name)


# ------- commands -----------------------

if __name__ == '__main__':   # upload file to backblaze
    # additional imports
    import argparse
    import os
    import sys
    from configparser import RawConfigParser

    def cp_hlp():
        cp_parser.print_help(sys.stderr)
        sys.exit(1)

    # parse commandline
    cp_parser = argparse.ArgumentParser(description=
                                        "Upload single file to backblaze b2 "
                                        "(note: used as module this has more functionalities).",
                                        formatter_class=argparse.RawTextHelpFormatter)
    cp_parser.add_argument('filename', action="store",
                           help="source filename include path; for b2 target basename is used, for path use --b2path")
    cp_parser.add_argument('-e', '--env', action="store", dest='envfile',
                           help=".ini file containing b2 parameters; default: use environment variables instead"
                           " (they must be prefixed BK_, ie. BK_B2_APP_KEY_ID and so on)")
    cp_parser.add_argument('-s', '--section', action="store", dest='section', default='backup',
                           help="section in --env .ini file; default: backup, ie. read the section [backup]\n"
                           "the section must contain these parameters:\n"
                           "\tB2_APP_KEY_ID=000xxxxxxxxxxxx000000000n\n"
                           "\tB2_APP_KEY=xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n"
                           "\tB2_BUCKET_NAME=bucketname")
    cp_parser.add_argument('-p', '--b2path', action="store", dest='b2path', default='',
                           help="destination path; default: none, ie. store file into bucket root")
    cp_par = cp_parser.parse_args()

    if cp_par.filename and os.path.isfile(cp_par.filename):
        # get parameters about the target bucket
        if cp_par.envfile and cp_par.section and os.path.isfile(cp_par.envfile):
            cp_config = RawConfigParser()
            cp_config.read(cp_par.envfile)
            cp_app_key_id = cp_config.get(cp_par.section, 'B2_APP_KEY_ID')
            cp_app_key = cp_config.get(cp_par.section, 'B2_APP_KEY')
            cp_bucket_name = cp_config.get(cp_par.section, 'B2_BUCKET_NAME')
        else:
            cp_app_key_id = os.environ.get('BK_B2_APP_KEY_ID')
            cp_app_key = os.environ.get('BK_B2_APP_KEY')
            cp_bucket_name = os.environ.get('BK_B2_BUCKET_NAME')

        if cp_app_key_id and cp_app_key and cp_bucket_name:
            # copy !
            cp_b2 = BackBlazeB2()
            cp_b2.authorize("production", cp_app_key_id, cp_app_key)
            cp_b2.set_bucket(cp_bucket_name)

            with open(cp_par.filename, 'rb') as f:
                cp_b2.upload_file(os.path.join(cp_par.b2path, os.path.basename(cp_par.filename)), f)
        else:
            print('FAILED: Parameters for the target bucket were not found, in env-variables or in --env file')
            print()
            cp_hlp()
    else:
        cp_hlp()
